﻿/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace EventsModel
{
	using System;
	using System.Collections.Generic;

	using MicrosoftResearch.Infer;
	using MicrosoftResearch.Infer.Distributions;
	using MicrosoftResearch.Infer.Maths;
	using MicrosoftResearch.Infer.Utils;
    using System.Globalization;
    using MicrosoftResearch.Infer.Models;
    using System.Diagnostics;

	class Program
    {
        /// <summary>
		/// The entry point for the program.
		/// </summary>
		/// <param name="args">The array of command-line arguments.</param>
		static void Main(string[] args)
        {
            Console.WriteLine("Starting...");
            System.Globalization.CultureInfo customCulture = (System.Globalization.CultureInfo)System.Threading.Thread.CurrentThread.CurrentCulture.Clone();
            customCulture.NumberFormat.NumberDecimalSeparator = ".";
            System.Threading.Thread.CurrentThread.CurrentCulture = customCulture;

            // load data
			Console.WriteLine("Loading data...");
            bool ignoreHeader;
            string datadir = @"..\..\data2\"; //windows
            //string datadir = @"../../data/"; //unix
            string datafile = datadir + @"Esplanade.tsv"; 
            Console.WriteLine("Dataset: " + datafile);
            int[] featuresToIgnore = new int[] { 0 };
            string[] dates;
            string[][] titles;
            double[] targets;
            Vector[] baseVectors;
            Vector[][] eventsVectors = DataUtils.Read(datafile, ignoreHeader = true, out baseVectors, out targets, out dates, out titles);

            // Build dataset
            DataSet dataset = new DataSet(baseVectors, eventsVectors, targets, dates, titles);
            //dataset = StandardizeDataset(dataset);
            //dataset = ShuffleDataset(dataset);
            Console.WriteLine("num instances=" + dataset.GetNumInstances());
            Console.WriteLine("num base features=" + dataset.GetNumBaseFeatures());
            Console.WriteLine("num events features=" + dataset.GetNumEventsFeatures());

            //Console.WriteLine("converted features=" + StringUtil.ArrayToString(baseVectors));
            System.IO.File.WriteAllText(datadir + @"baseFeaturesDebug.txt", StringUtil.ArrayToString(baseVectors));
            //Console.WriteLine("events features=" + StringUtil.ArrayToString(eventsFeatureVectors));
            System.IO.File.WriteAllText(datadir + @"eventsFeaturesDebug.txt", StringUtil.ArrayToString(eventsVectors));

            EventsLRModel eventsModel = new EventsLRModel();
            Console.WriteLine("computing model evidence...");
            eventsModel.CreateModelTraining(dataset.GetNumBaseFeatures(), dataset.GetNumEventsFeatures());
            double logEvidence = eventsModel.ComputeModelEvidence(dataset);
            Console.WriteLine("log evidence=" + logEvidence);
            Console.WriteLine("estimating model parameters...");
            EventsModelData weightsPosteriors = eventsModel.InferModelWeights(dataset);
            System.IO.File.WriteAllText(datadir + @"eventsWeightsDebug.txt", weightsPosteriors.eventsWeightsPosterior.ToString());
            System.IO.File.WriteAllText(datadir + @"dummyWeightsDebug.txt", weightsPosteriors.dummyWeightsPosterior.ToString());
            Console.WriteLine("estimating latent contributions...");
            eventsModel.CreateModelTesting(dataset.GetNumBaseFeatures(), dataset.GetNumEventsFeatures());
            /*EventsModelData latentPosteriors = eventsModel.InferLatentContributions(weightsPosteriors, dataset);
            // output latent contributions to file
            using (System.IO.StreamWriter file = new System.IO.StreamWriter(datafile.Replace(".tsv", "_contributions.tsv")))
            {
                file.WriteLine("stop_id,date,start_time,end_time,true_total_arrivals,dummy_contribution,events_contributions");
                for (int i = 0; i < dataset.targets.Length; i++)
                {
                    string dummyStr = "N(" + latentPosteriors.latentDummyContributionsPosterior[i].GetMean() + "; " + latentPosteriors.latentDummyContributionsPosterior[i].GetVariance() + ")";
                    string eventsStr = "";
                    for(int j=0; j<latentPosteriors.latentEventsContributionsPosterior[i].Count; j++) 
                    {
                        eventsStr += dataset.titles[i][j].Replace("|","") + " N(" + latentPosteriors.latentEventsContributionsPosterior[i][j].GetMean() + "; " + latentPosteriors.latentEventsContributionsPosterior[i][j].GetVariance()+") | ";
                    }
                    file.WriteLine(dataset.dates[i] + "," + dataset.targets[i] + "," + dummyStr + "," + eventsStr);
                }
                file.Flush();
                file.Close();
            }*/
            DistributionStructArray<Gaussian, double> predictions = eventsModel.MakePredictions(weightsPosteriors, dataset);
            double[] preds = ComputeError(dataset.targets, predictions);

			Console.WriteLine("Press key to quit");
			Console.ReadKey();
		}

        public static double[] ComputeError(double[] targets, DistributionStructArray<Gaussian, double> yPosterior)
        {
            Array predictiveDists = yPosterior.ToArray();
            double[] preds = new double[targets.Length];
            for (int i = 0; i < targets.Length; i++)
            {
                Gaussian predictiveDist = (Gaussian)predictiveDists.GetValue(i);
                //Console.WriteLine(targets[i] + " - " + predictiveDist);
                preds[i] = predictiveDist.GetMean();
            }
            return ComputeError(targets, preds);
        }

        public static double[] ComputeError(double[] targets, double[] preds)
        {

            double meanTarget = MathNet.Numerics.Statistics.ArrayStatistics.Mean(targets);
            double sumError = 0;
            double sumSquaredError = 0;
            double sumNorms = 0;
            double sumNormsSquared = 0;
            for (int i = 0; i < targets.Length; i++)
            {
                double error = targets[i] - preds[i];
                sumError += Math.Abs(error);
                sumSquaredError += Math.Pow(error, 2);
                sumNorms += Math.Abs(targets[i] - meanTarget);
                sumNormsSquared += Math.Pow(targets[i] - meanTarget, 2);
            }
            double corr = MathNet.Numerics.Statistics.Correlation.Pearson(targets, preds);
            double mae = sumError / targets.Length;
            double rmse = Math.Sqrt(sumSquaredError / targets.Length);
            double rae = sumError / sumNorms;
            double rrse = sumSquaredError / sumNormsSquared;
            Console.WriteLine("\nGLOBAL ERROR STATISTICS:");
            Console.WriteLine("corr=" + (corr * 100).ToString("0.000") + "\tmae=" + mae.ToString("0.000") + "\trmse=" + rmse.ToString("0.000") + "\trae=" + (rae * 100).ToString("0.000") + "\trrse=" + (rrse * 100).ToString("0.000"));
            Console.WriteLine("TSV:" + (corr * 100).ToString("0.000") + "\t" + mae.ToString("0.000") + "\t" + rmse.ToString("0.000") + "\t" + (rae * 100).ToString("0.000") + "\t" + (rrse * 100).ToString("0.000"));
            return preds;
        }

        public static void Shuffle<T>(Random rng, T[] array)
        {
            int n = array.Length;
            while (n > 1)
            {
                int k = rng.Next(n--);
                T temp = array[n];
                array[n] = array[k];
                array[k] = temp;
            }
        }

        public static void SplitArray<T>(T[] source, int index, out T[] first, out T[] second)
        {
            int len2 = source.Length - index;
            first = new T[index];
            second = new T[len2];
            Array.Copy(source, 0, first, 0, index);
            Array.Copy(source, index, second, 0, len2);
        }
	}
}