﻿using MicrosoftResearch.Infer;
using MicrosoftResearch.Infer.Distributions;
using MicrosoftResearch.Infer.Maths;
using MicrosoftResearch.Infer.Models;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;

namespace EventsModel
{
    public class EventsLRModel
    {
        public int numBaseFeatures;
        public int numEventsFeatures;

        public InferenceEngine trainInferenceEngine;
        public Variable<bool> evidence;
        public Variable<int> numTraining;
        public Range Ntrain;
        public Variable<Vector> dummyWeights;
        public VariableArray<Vector> baseFeaturesTrain;
        public VariableArray<double> dummyContributionsTrain;
        public VariableArray<int> numEventsTrain;
        public Range Etrain;
        public Variable<Vector> eventsFeatureWeights;
        public VariableArray<VariableArray<Vector>, Vector[][]> eventsFeaturesTrain;
        public VariableArray<VariableArray<double>, double[][]> eventsContributionsTrain;
        public VariableArray<double> observedCountsTrain;

        public InferenceEngine testInferenceEngine;
        public Variable<int> numTesting;
        public Range Ntest;
        public Variable<VectorGaussian> dummyWeightsPosterior;
        public VariableArray<Vector> baseFeaturesTest;
        public VariableArray<double> dummyContributionsTest;
        public VariableArray<int> numEventsTest;
        public Range Etest;
        public Variable<VectorGaussian> eventsFeatureWeightsPosterior;
        public VariableArray<VariableArray<Vector>, Vector[][]> eventsFeaturesTest;
        public VariableArray<VariableArray<double>, double[][]> eventsContributionsTest;
        public VariableArray<double> observedCountsTest;

        public void CreateModelTraining(int numBaseFeatures, int numEventsFeatures)
        {
            this.numBaseFeatures = numBaseFeatures;
            this.numEventsFeatures = numEventsFeatures;

            evidence = Variable.Bernoulli(0.5).Named("evidence");
            IfBlock block = Variable.If(evidence);

            // start of model 

            numTraining = Variable.New<int>().Named("numTraining");
            Ntrain = new Range(numTraining).Named("Ntrain");

            dummyWeights = Variable.VectorGaussianFromMeanAndPrecision(Vector.Zero(numBaseFeatures), PositiveDefiniteMatrix.Identity(numBaseFeatures)).Named("dummyWeights");
            
            baseFeaturesTrain = Variable.Array<Vector>(Ntrain).Named("baseFeaturesTrain");
            dummyContributionsTrain = Variable.Array<double>(Ntrain).Named("dummyContributionsTrain");

            numEventsTrain = Variable.Array<int>(Ntrain).Named("numEventsTrain");
            Etrain = new Range(numEventsTrain[Ntrain]).Named("Etrain");
            eventsFeatureWeights = Variable.VectorGaussianFromMeanAndPrecision(Vector.Zero(numEventsFeatures), PositiveDefiniteMatrix.Identity(numEventsFeatures)).Named("eventsFeatureWeights");
            eventsFeaturesTrain = Variable.Array(Variable.Array<Vector>(Etrain), Ntrain).Named("eventsFeaturesTrain");
            eventsContributionsTrain = Variable.Array(Variable.Array<double>(Etrain), Ntrain).Named("eventsContributionsTrain");
            
            observedCountsTrain = Variable.Array<double>(Ntrain).Named("observedCountsTrain");

            Variable<double> noise = Variable.GaussianFromMeanAndVariance(0, 10);

            using (Variable.ForEach(Ntrain))
            {
                dummyContributionsTrain[Ntrain] = Variable.GaussianFromMeanAndVariance(Variable.InnerProduct(baseFeaturesTrain[Ntrain], dummyWeights), 5.0);
                Variable.ConstrainPositive(dummyContributionsTrain[Ntrain]); 
                using (Variable.ForEach(Etrain))
                {
                    eventsContributionsTrain[Ntrain][Etrain] = Variable.GaussianFromMeanAndVariance(Variable.InnerProduct(eventsFeaturesTrain[Ntrain][Etrain], eventsFeatureWeights), 5.0);
                    Variable.ConstrainPositive(eventsContributionsTrain[Ntrain][Etrain]);
                }
                observedCountsTrain[Ntrain] = dummyContributionsTrain[Ntrain] + Variable.Sum(eventsContributionsTrain[Ntrain]);// +noise;
            }

            // end of model 

            block.CloseBlock();

            if (trainInferenceEngine == null)
            {
                trainInferenceEngine = new InferenceEngine(new ExpectationPropagation());
                trainInferenceEngine.NumberOfIterations = 20;
                Ntrain.AddAttribute(new Sequential());  // needed to get stable convergence 
                trainInferenceEngine.Compiler.UseSerialSchedules = true; 
            }
        }
        
        public void CreateModelTesting(int numBaseFeatures, int numEventsFeatures)
        {
            this.numBaseFeatures = numBaseFeatures;
            this.numEventsFeatures = numEventsFeatures;

            IfBlock block = Variable.If(evidence);

            // start of model 

            numTesting = Variable.New<int>().Named("numTesting");
            Ntest = new Range(numTesting).Named("Ntest");

            dummyWeightsPosterior = Variable.New<VectorGaussian>().Named("dummyWeightsPosterior");
            Variable<Vector> testDummyWeights = Variable<Vector>.Random(dummyWeightsPosterior);

            baseFeaturesTest = Variable.Array<Vector>(Ntest).Named("baseFeaturesTest");
            dummyContributionsTest = Variable.Array<double>(Ntest).Named("dummyContributionsTest");

            numEventsTest = Variable.Array<int>(Ntest).Named("numEventsTest");
            Etest = new Range(numEventsTest[Ntest]).Named("Etest");
            eventsFeatureWeightsPosterior = Variable.New<VectorGaussian>().Named("eventsFeatureWeightsPosterior");
            Variable<Vector> testEventsWeights = Variable<Vector>.Random(eventsFeatureWeightsPosterior);
            eventsFeaturesTest = Variable.Array(Variable.Array<Vector>(Etest), Ntest).Named("eventsFeaturesTest");
            eventsContributionsTest = Variable.Array(Variable.Array<double>(Etest), Ntest).Named("eventsContributionsTest");

            observedCountsTest = Variable.Array<double>(Ntest).Named("observedCountsTest");

            Variable<double> noise = Variable.GaussianFromMeanAndVariance(0, 10);

            using (Variable.ForEach(Ntest))
            {
                dummyContributionsTest[Ntest] = Variable.GaussianFromMeanAndVariance(Variable.InnerProduct(baseFeaturesTest[Ntest], testDummyWeights), 5.0);
                Variable.ConstrainPositive(dummyContributionsTest[Ntest]);
                using (Variable.ForEach(Etest))
                {
                    eventsContributionsTest[Ntest][Etest] = Variable.GaussianFromMeanAndVariance(Variable.InnerProduct(eventsFeaturesTest[Ntest][Etest], testEventsWeights), 5.0);
                    Variable.ConstrainPositive(eventsContributionsTest[Ntest][Etest]);
                }
                observedCountsTest[Ntest] = dummyContributionsTest[Ntest] + Variable.Sum(eventsContributionsTest[Ntest]);// +noise;
            }

            // end of model 

            block.CloseBlock();

            if (testInferenceEngine == null)
            {
                testInferenceEngine = new InferenceEngine(new ExpectationPropagation());
                testInferenceEngine.NumberOfIterations = 20;
                Ntest.AddAttribute(new Sequential());  // needed to get stable convergence 
                testInferenceEngine.Compiler.UseSerialSchedules = true;
            }
        }

        public EventsModelData InferModelWeights(DataSet dataset)
        {
            int[] numEventsPerInstance = new int[dataset.GetNumInstances()];
            for (int i = 0; i < dataset.eventsFeatureVectors.Length; i++)
            {
                numEventsPerInstance[i] = dataset.eventsFeatureVectors[i].Length;
            }
            evidence.ObservedValue = true;
            numTraining.ObservedValue = dataset.GetNumInstances();
            numEventsTrain.ObservedValue = numEventsPerInstance;
            baseFeaturesTrain.ObservedValue = dataset.baseFeatureVectors;
            eventsFeaturesTrain.ObservedValue = dataset.eventsFeatureVectors;
            observedCountsTrain.ObservedValue = dataset.targets;

            VectorGaussian dummyWeightsPosterior = trainInferenceEngine.Infer<VectorGaussian>(dummyWeights);
            VectorGaussian eventsWeightsPosterior = trainInferenceEngine.Infer<VectorGaussian>(eventsFeatureWeights);
            //Console.WriteLine("Posterior over the routineWeights: " + Environment.NewLine + routineWeightsPosterior);
            //Console.WriteLine("Posterior over the eventsWeights: " + Environment.NewLine + eventsWeightsPosterior);
            return new EventsModelData(dummyWeightsPosterior, eventsWeightsPosterior);
        }

        public double ComputeModelEvidence(DataSet dataset)
        {
            int[] numEventsPerInstance = new int[dataset.GetNumInstances()];
            for (int i = 0; i < dataset.eventsFeatureVectors.Length; i++)
            {
                numEventsPerInstance[i] = dataset.eventsFeatureVectors[i].Length;
            }
            numTraining.ObservedValue = dataset.GetNumInstances();
            numEventsTrain.ObservedValue = numEventsPerInstance;
            baseFeaturesTrain.ObservedValue = dataset.baseFeatureVectors;
            eventsFeaturesTrain.ObservedValue = dataset.eventsFeatureVectors;
            observedCountsTrain.ObservedValue = dataset.targets;

            double logEvidence = trainInferenceEngine.Infer<Bernoulli>(evidence).LogOdds;
            return logEvidence;
        }

        public EventsModelData InferLatentContributions(EventsModelData posteriors, DataSet dataset)
        {
            int[] numEventsPerInstance = new int[dataset.GetNumInstances()];
            for (int i = 0; i < dataset.eventsFeatureVectors.Length; i++)
            {
                numEventsPerInstance[i] = dataset.eventsFeatureVectors[i].Length;
            }
            evidence.ObservedValue = true;
            numTesting.ObservedValue = dataset.GetNumInstances();
            numEventsTest.ObservedValue = numEventsPerInstance;
            baseFeaturesTest.ObservedValue = dataset.baseFeatureVectors;
            eventsFeaturesTest.ObservedValue = dataset.eventsFeatureVectors;
            observedCountsTest.ObservedValue = dataset.targets;
            dummyWeightsPosterior.ObservedValue = posteriors.dummyWeightsPosterior;
            eventsFeatureWeightsPosterior.ObservedValue = posteriors.eventsWeightsPosterior;

            DistributionStructArray<Gaussian, double> latentDummyContributionsPosterior = testInferenceEngine.Infer<DistributionStructArray<Gaussian, double>>(dummyContributionsTest);
            DistributionRefArray<DistributionStructArray<Gaussian, double>, double[]> latentEventsContributionsPosterior = testInferenceEngine.Infer<DistributionRefArray<DistributionStructArray<Gaussian, double>, double[]>>(eventsContributionsTest);
            //Console.WriteLine("Posterior over the latentRoutineContributionsPosterior: " + Environment.NewLine + latentRoutineContributionsPosterior);
            //Console.WriteLine("Posterior over the latentEventsContributionsPosterior: " + Environment.NewLine + latentEventsContributionsPosterior);
            return new EventsModelData(latentDummyContributionsPosterior, latentEventsContributionsPosterior);
        }

        public DistributionStructArray<Gaussian, double> MakePredictions(EventsModelData posteriors, DataSet dataset)
        {
            int[] numEventsPerInstance = new int[dataset.GetNumInstances()];
            for (int i = 0; i < dataset.eventsFeatureVectors.Length; i++)
            {
                numEventsPerInstance[i] = dataset.eventsFeatureVectors[i].Length;
            }
            evidence.ObservedValue = true;
            numTesting.ObservedValue = dataset.GetNumInstances();
            numEventsTest.ObservedValue = numEventsPerInstance;
            baseFeaturesTest.ObservedValue = dataset.baseFeatureVectors;
            eventsFeaturesTest.ObservedValue = dataset.eventsFeatureVectors;
            //observedCountsTest.ObservedValue = dataset.targets;
            dummyWeightsPosterior.ObservedValue = posteriors.dummyWeightsPosterior;
            eventsFeatureWeightsPosterior.ObservedValue = posteriors.eventsWeightsPosterior;

            DistributionStructArray<Gaussian, double> observedCountsPosterior = testInferenceEngine.Infer<DistributionStructArray<Gaussian, double>>(observedCountsTest);
            return observedCountsPosterior;
        }
    }

    public class EventsModelData
    {
        public VectorGaussian dummyWeightsPosterior;
        public VectorGaussian eventsWeightsPosterior;
        public DistributionStructArray<Gaussian, double> latentDummyContributionsPosterior;
        public DistributionRefArray<DistributionStructArray<Gaussian, double>, double[]> latentEventsContributionsPosterior;

        public EventsModelData(VectorGaussian dummyWeightsPosterior, VectorGaussian eventsWeightsPosterior)
        {
            this.dummyWeightsPosterior = dummyWeightsPosterior;
            this.eventsWeightsPosterior = eventsWeightsPosterior;
            this.latentDummyContributionsPosterior = null;
            this.latentEventsContributionsPosterior = null;
        }

        public EventsModelData(DistributionStructArray<Gaussian, double> latentDummyContributionsPosterior, DistributionRefArray<DistributionStructArray<Gaussian, double>, double[]> latentEventsContributionsPosterior)
        {
            this.dummyWeightsPosterior = null;
            this.eventsWeightsPosterior = null;
            this.latentDummyContributionsPosterior = latentDummyContributionsPosterior;
            this.latentEventsContributionsPosterior = latentEventsContributionsPosterior;
        }
    }

    public class DataSet
    {
        public Vector[] baseFeatureVectors;
        public Vector[][] eventsFeatureVectors;
        public double[] targets;
        public string[] dates;
        public string[][] titles;

        public DataSet()
        {
        }

        public DataSet(Vector[] baseFeatureVectors, Vector[][] eventsFeatureVectors, double[] targets, string[] dates, string[][] titles)
        {
            Debug.Assert(targets.Length == baseFeatureVectors.Length);
            Debug.Assert(targets.Length == eventsFeatureVectors.Length);
            Debug.Assert(targets.Length == dates.Length);
            Debug.Assert(targets.Length == titles.Length);
            this.baseFeatureVectors = baseFeatureVectors;
            this.eventsFeatureVectors = eventsFeatureVectors;
            this.targets = targets;
            this.dates = dates;
            this.titles = titles;
        }

        public int GetNumInstances()
        {
            return targets.Length;
        }

        public int GetNumBaseFeatures()
        {
            return baseFeatureVectors[0].Count;
        }

        public int GetNumEventsFeatures()
        {
            for (int i = 0; i < eventsFeatureVectors.Length; i++)
            {
                if (eventsFeatureVectors[i].Length > 0)
                {
                    return eventsFeatureVectors[i][0].Count;
                }
            }
            return 0;
        }
    }
}
